# -*- coding: utf-8 -*-
"""
Created on Mon Feb 24 09:26:21 2025

@author: mstep
"""

def rectify(image, points):
    '''Reshape the image so the points form a rectangle.'''
    

#%%

import math

def dist(p1, p2):
    dx = p2[0] - p1[0]    
    dy = p2[1] - p1[1]    
    return math.sqrt(dx * dx + dy * dy)

#%%
import tkinter as tk
from PIL import ImageTk, Image
from tkinter import messagebox, filedialog
import cv2
import numpy as np

class RectifyApp:
    def __init__(self):
        self.original_pil_image = None
        self.scale_factor = 1
        self.corners = []

        self.window = tk.Tk()
        self.window.title('rectify_trapezoid')
        self.window.protocol('WM_DELETE_WINDOW', self.kill_callback)
        self.window.geometry('800x600')

        # Build the menus.
        self.menubar = tk.Menu(self.window)

        self.menu_file = tk.Menu(self.menubar, tearoff=False)
        self.menubar.add_cascade(label='File', menu=self.menu_file)
        self.menu_file.add_command(label='Open...', command=self.open_file,
            accelerator='Ctrl+O')
        self.menu_file.add_command(label='Save As...', command=self.save_as,
            accelerator='Ctrl+S')
        self.menu_file.add_separator()
        self.menu_file.add_command(label='Reset', command=self.reset,
            accelerator='Ctrl+R')
        self.menu_file.add_command(label='Exit', command=self.kill_callback)

        self.menu_scale = tk.Menu(self.menubar, tearoff=False)
        self.menubar.add_cascade(label='Scale', menu=self.menu_scale)
        self.menu_scale.add_command(label='100%', command=lambda: self.scale(1.00))
        self.menu_scale.add_command(label='75%', command=lambda: self.scale(0.75))
        self.menu_scale.add_command(label='50%', command=lambda: self.scale(0.50))
        self.menu_scale.add_command(label='25%', command=lambda: self.scale(0.25))
        self.menu_scale.add_command(label='15%', command=lambda: self.scale(0.15))

        # Set the main menu.
        self.window.config(menu=self.menubar)

        # Disable initially disabled menus.
        self.disable_menus()

        # Build the window surface.
        self.canvas = tk.Canvas(self.window, bg='white', borderwidth=2,
            relief=tk.SUNKEN, cursor='tcross')
        self.canvas.pack(padx=10, pady=(0, 10), side=tk.BOTTOM, fill=tk.BOTH,
            expand=True)

        # Bind menu accelerators.
        self.window.bind_all('<Control-o>', lambda event: self.open_file())
        self.window.bind_all('<Control-s>', lambda event: self.save_as())
        self.window.bind_all('<Control-r>', lambda event: self.reset())
        self.canvas.bind("<Button-1>", self.mouse_down)

        # Display the window.
        self.window.focus_force()
        self.window.mainloop()

    def kill_callback(self):
        # Destroy the tkinter window.
        self.window.destroy()

    def scale(self, scale_factor):
        self.scale_factor = scale_factor
        self.show_current_image()

    def disable_menus(self):
        # Disable menus that should not be enabled while no image is loaded.
        self.menu_file.entryconfig('Save As...', state='disabled')
        self.menu_file.entryconfig('Reset', state='disabled')

    def enable_menus(self):
        # Enable menus that should be enabled when an image is loaded.
        self.menu_file.entryconfig('Save As...', state='normal')
        self.menu_file.entryconfig('Reset', state='normal')

    def show_current_image(self):
        if self.current_pil_image == None: return

        # Resize the image.
        wid = int(self.scale_factor * self.current_pil_image.width)
        hgt = int(self.scale_factor * self.current_pil_image.height)
        resized_image = self.current_pil_image.resize((wid, hgt))

        # Display the image.
        self.current_tk_image = ImageTk.PhotoImage(resized_image)
        self.canvas.delete('all')
        self.display_image = self.canvas.create_image(0, 0, anchor=tk.NW,
            image=self.current_tk_image)

        # Draw any defined corners.
        for corner in self.corners:
            dx = dy = 4
            x = self.scale_factor * corner[0]
            y = self.scale_factor * corner[1]
            self.canvas.create_rectangle(
                x - dx, y - dy,
                x + dx, y + dy,
                outline='black', fill='white')

    # File menu.
    def open_file(self):
        file_types = [('Image Files', '.png .jpg .bmp .gif')]
        filename = tk.filedialog.askopenfilename(
            defaultextension='.png', filetypes=file_types,
            initialdir='.', title='Open File')
        if not filename: return
        try:
            self.original_pil_image = Image.open(filename)

            # Use RGB mode.
            if self.original_pil_image.mode != 'RGB':
                self.original_pil_image = self.original_pil_image.convert(mode='RGB')

            self.current_pil_image = self.original_pil_image.copy()
            self.show_current_image()
            self.enable_menus()
        except Exception as e:
            messagebox.showinfo('Open Error', f'Error opening image.\n{e}')
        self.filename = filename

    def save_as(self):
        file_types = [('Image Files', '.png .jpg .bmp .gif')]
        filename = filedialog.asksaveasfilename(
            defaultextension='.png', filetypes=file_types,
            initialdir='.', title='Save As')
        if not filename: return
        try:
            self.current_pil_image.save(filename)
        except Exception as e:
            messagebox.showinfo('Save Error', f'Error saving image.\n{e}')

    def reset(self):
        if self.original_pil_image == None: return
        self.current_pil_image = self.original_pil_image.copy()
        self.show_current_image()

    def mouse_down(self, event):
        # If we already have 4 corners, reset the list.
        if len(self.corners) == 4:
            self.corners = []

        # Get the new corner, scaling if necessary.
        x = event.x / self.scale_factor
        y = event.y / self.scale_factor
        self.corners.append([x, y])

        # Redisplay to show the new corner.
        self.show_current_image()

        # If we now have 4 corners, rectify.
        if len(self.corners) == 4:
            self.rectify_trapezoid()

    def rectify_trapezoid(self):
        image = cv2.imread(self.filename)
        if image is None:
            messagebox.showinfo('File Error', f'Cannot find file {self.filename}.')
            return

        # Get the result rectangle's dimensions.
        xmin = min([p[0] for p in self.corners])
        xmax = max([p[0] for p in self.corners])
        ymin = min([p[1] for p in self.corners])
        ymax = max([p[1] for p in self.corners])
        wid = xmax - xmin
        hgt = ymax - ymin

        # Make margins 10% of width/height.
        margin_x = int(wid * 0.1)
        margin_y = int(hgt * 0.1)

        # Calculate the horizontal distance from xmin to the first point.
        dx1 = self.corners[0][0] - xmin
        dx2 = xmax - self.corners[1][0]
        dx = (dx1 + dx2) / 2

        pts1 = np.float32(self.corners) # Source points
        pts2 = np.float32([
            [margin_x + dx, margin_y],
            [margin_x + wid - dx, margin_y],
            [margin_x + wid, margin_y + hgt],
            [margin_x, margin_y + hgt]]) # Destination points

        # Get the transformation matrix.
        matrix = cv2.getPerspectiveTransform(pts1, pts2)

        # Get output image dimensions.
        result_wid = int(wid + 2 * margin_x)
        result_hgt = int(hgt + 2 * margin_y)

        # Transform.
        new_image = cv2.warpPerspective(image,
            matrix, (result_wid, result_hgt))

        # Convert colors from BGR to RGB.
        new_image = cv2.cvtColor(new_image, cv2.COLOR_BGR2RGB)

        # Remove the current corners.
        self.corners = []

        # Convert to PIL image and save.
        self.current_pil_image = Image.fromarray(new_image)
        self.show_current_image()

#%%

app = RectifyApp()
del app
